package org.certificatetransparency.ctlog;

import org.certificatetransparency.ctlog.PrecertChainEntry;
import org.certificatetransparency.ctlog.X509ChainEntry;


public class LogEntry {
  public X509ChainEntry x509Entry;
  public PrecertChainEntry precertEntry;
}
