export INSTANCE_TYPE="mirror"
export PROJECT="ct-mirror-certly1"
export REGION="us-central1"
export CLUSTER="ct-mirror-certly1"
export ZONES="a c f"
export MIRROR_NUM_REPLICAS_PER_ZONE=2
export MIRROR_TARGET_URL="https://log.certly.io"
export MIRROR_TARGET_PUBLIC_KEY="certly_1.pem"
export MONITORING="gcm"
