export INSTANCE_TYPE="mirror"
export PROJECT="ct-mirror-izenpe1"
export REGION="us-central1"
export CLUSTER="ct-mirror-izenpe1"
export ZONES="a c f"
export MIRROR_NUM_REPLICAS_PER_ZONE=2
export MIRROR_TARGET_URL="https://ct.izenpe.com"
export MIRROR_TARGET_TLS_VERSION="tlsv1"
export MIRROR_TARGET_PUBLIC_KEY="izenpe_1.pem"
export MONITORING="gcm"
