<html>
  <head>
    <title>Certificate Transparency dashboard</title>
  </head>
  <body>
    <H1>Welcome to the Certificate Transparency dashboard!</H1>
    <div id="pages">Try the dashboard out at %(pages)s</div>
    <br>
    <div id="time">%(current_time)s</div>
  </body>
</html>
