export INSTANCE_TYPE="mirror"
export PROJECT="ct-mirror-submariner"
export REGION="us-east1"
export CLUSTER="ct-mirror-submariner"
export ZONES="c d b"
export MIRROR_NUM_REPLICAS_PER_ZONE=2
export MIRROR_TARGET_URL="https://ct.googleapis.com/submariner"
export MIRROR_TARGET_PUBLIC_KEY="submariner.pem"
export MONITORING="gcm"
